package com.yunzhi.shigongli.service.impl;

import cn.binarywang.wx.miniapp.api.WxMaService;
import cn.binarywang.wx.miniapp.bean.WxMaSubscribeMessage;
import com.github.binarywang.wxpay.bean.notify.WxPayOrderNotifyResult;
import com.github.binarywang.wxpay.bean.notify.WxPayRefundNotifyResult;
import com.github.binarywang.wxpay.bean.order.WxPayMpOrderResult;
import com.github.binarywang.wxpay.constant.WxPayConstants;
import com.yunzhi.shigongli.common.*;
import com.yunzhi.shigongli.entity.*;
import com.yunzhi.shigongli.event.PaySuccessEvent;
import com.yunzhi.shigongli.event.RefundSuccessEvent;
import com.yunzhi.shigongli.mapper.*;
import com.yunzhi.shigongli.service.ITaOrderPayService;

import lombok.extern.slf4j.Slf4j;
import me.chanjar.weixin.common.error.WxErrorException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * <p>
 * 订单支付 服务实现类
 * </p>
 *
 * @author yansen
 * @since 2021-08-17
 */
@Slf4j
@Service
public class TaOrderPayServiceImpl extends BaseServiceImpl<TaOrderPayMapper, TaOrderPay> implements ITaOrderPayService {

    @Autowired
    WxPayUtils wxPayUtils;

    @Autowired
    TaOrderPayMapper orderPayMapper;

    @Autowired
    TaOrderMapper taOrderMapper;

    @Autowired
    TaOrderSubMapper taOrderSubMapper;

    @Autowired
    ApplicationEventPublisher eventPublisher;

    @Autowired
    TaSequenceMapper taSequenceMapper;

    @Autowired
    TaShopPackageMapper shopPackageMapper;

    @Autowired
    TaPersonMapper taPersonMapper;

    @Autowired
    TaVerifyTargetMapper taVerifyTargetMapper;

    @Autowired
    TaPersonAccountMapper personAccountMapper;

    @Autowired
    TaPersonAccountLogMapper accountLogMapper;

    @Autowired
    TaShopAccountMapper shopAccountMapper;

    @Autowired
    TaShopAccountLogMapper shopAccountLogMapper;

    @Autowired
    TaMiniappMessageTplMapper messageTplMapper;

    @Autowired
    WxUtils wxUtils;

    /**
     * 支付回调
     * @param notifyResult
     * @throws Exception
     */
    @Override
    public void notifyOrder(WxPayOrderNotifyResult notifyResult) throws Exception {
        TaOrderPay orderPay = getById(notifyResult.getOutTradeNo());
        String orderId = orderPay.getOrderId();
        String orderSubId = orderPay.getSubOrderId();

        TaOrder taOrder = taOrderMapper.selectById(orderId);
        TaOrderSub taOrderSub = taOrderSubMapper.selectById(orderSubId);
        TaPerson taPerson = taPersonMapper.selectById(orderPay.getPersonId());

        // 如果不是待支付，说明已经处理过了
        if (null != taOrderSub) {
            if (Constants.ORDER_READY != taOrderSub.getStatus()) {
                return;
            }
        } else {
            if (Constants.ORDER_READY != taOrder.getStatus()) {
                return;
            }
        }

        boolean isSuccess = WxPayConstants.ResultCode.SUCCESS.equals(notifyResult.getResultCode());

        orderPay.setPayDate(DateUtils.from(notifyResult.getTimeEnd(), "yyyyMMddHHmmss"));
        orderPay.setStatus(isSuccess ? Constants.PAY_SUCCESS : Constants.PAY_FAIL);
        orderPay.setMessage(isSuccess ? null : notifyResult.getErrCodeDes());

        orderPayMapper.updateById(orderPay);

        if (isSuccess) {
            // 更新订单为已支付
            if (!StringUtils.isEmpty(orderSubId)) {
                taOrderSubMapper.updateStatus(orderSubId, Constants.ORDER_PAYED);
            } else {
                taOrderMapper.updateStatus(orderId, Constants.ORDER_PAYED);
                taOrderSubMapper.updateStatusByOrder(orderId, Constants.ORDER_PAYED);
            }

            List<TaOrderSub> orderSubList = new ArrayList<>();
            if (null != taOrder) {
                orderSubList = taOrderSubMapper.getListByOrder(taOrder.getOrderId());
            } else {
                orderSubList.add(taOrderSub);
            }

            // 商户累计营业额
            dealSaleCharges(orderSubList, false);

            // 发送核销码
            generateVerfiedRecords(taPerson, orderSubList);

            // 返现
            processCallback(orderPay, taPerson, orderSubList, taOrder, taOrderSub);

            PaySuccessEvent evt = new PaySuccessEvent(this, orderPay);
            eventPublisher.publishEvent(evt);
        }
    }

    private void dealSaleCharges(List<TaOrderSub> orderSubList, boolean isRefund) throws Exception {
        LocalDateTime now = LocalDateTime.now();

        for (TaOrderSub orderSub : orderSubList) {
            TaShopAccount account = shopAccountMapper.selectByShop(orderSub.getShopId());
            if (null == account) {
                log.error("商铺 {} 没有账户信息", orderSub.getShopId());
                continue;
            }

            Integer charge = orderSub.getCharges();
            shopAccountMapper.updateCharges(account.getAccountId(), "sales_charges", isRefund ? -charge : charge);

            TaShopAccountLog taShopAccountLog = new TaShopAccountLog();
            taShopAccountLog.setAccountId(account.getAccountId());
            taShopAccountLog.setShopId(orderSub.getShopId());
            taShopAccountLog.setSignType(isRefund ? Constants.ACCOUNT_EXPENSE : Constants.ACCOUNT_INCOME);
            taShopAccountLog.setAmountType(isRefund ? Constants.ACCOUNT_REFUND : Constants.ACCOUNT_SALES);
            // 退款金额 = 订单金额 - 返现金额
            Integer refundCharge = orderSub.getCharges() - orderSub.getCashback();
            taShopAccountLog.setAmount(isRefund ? -refundCharge : charge);
            taShopAccountLog.setTargetType(Constants.TARGET_ORDER_SUB);
            taShopAccountLog.setTargetId(orderSub.getSubOrderId());
            taShopAccountLog.setOrderId(orderSub.getOrderId());
            taShopAccountLog.setSubOrderId(orderSub.getSubOrderId());
            taShopAccountLog.setCreateDate(now);
            shopAccountLogMapper.insert(taShopAccountLog);

            if (!isRefund) {
                // 记录返现金额
                shopAccountMapper.updateCharges(account.getAccountId(), "cashback", orderSub.getCashback());

                taShopAccountLog = new TaShopAccountLog();
                taShopAccountLog.setAccountId(account.getAccountId());
                taShopAccountLog.setShopId(orderSub.getShopId());
                taShopAccountLog.setSignType(Constants.ACCOUNT_EXPENSE);
                taShopAccountLog.setAmountType(Constants.ACCOUNT_CASHBACK);
                taShopAccountLog.setAmount(-orderSub.getCashback());
                taShopAccountLog.setTargetType(Constants.TARGET_ORDER_SUB);
                taShopAccountLog.setTargetId(orderSub.getSubOrderId());
                taShopAccountLog.setOrderId(orderSub.getOrderId());
                taShopAccountLog.setSubOrderId(orderSub.getSubOrderId());
                taShopAccountLog.setCreateDate(now);
                shopAccountLogMapper.insert(taShopAccountLog);
            }
        }
    }

    private void processCallback(TaOrderPay originPay, TaPerson taPerson, List<TaOrderSub> orderSubList, TaOrder taOrder, TaOrderSub taOrderSub) throws Exception {
        if (null == orderSubList || orderSubList.size() < 1) {
            throw new Exception("未找到有效的订单项目记录");
        }

        Integer cashback = orderSubList.stream().mapToInt(TaOrderSub::getCashback).reduce(Integer::sum).getAsInt();
        if (cashback <= 0) {
            return ;
        }

        TaOrderPay orderPay = null != taOrder ? createByOrder(taOrder) : createByOrderSub(taOrderSub, taPerson);
        orderPay.setDescription("返现");
        orderPay.setIsRefund(true);
        orderPay.setCharges(cashback);
        orderPayMapper.insert(orderPay);

        wxPayUtils.refund(originPay, orderPay);
    }

    /**
     * 发放核销码
     * @param taPerson
     * @param orderSubList
     * @throws Exception
     */
    private void generateVerfiedRecords(TaPerson taPerson, List<TaOrderSub> orderSubList) throws Exception {
        LocalDateTime now = LocalDateTime.now();

        // 循环发放
        for (TaOrderSub item : orderSubList) {
            if (null == item) continue;

            // 已经处理过的, 则不发放
            int cnt = taVerifyTargetMapper.getCountBy(item.getOrderId(), item.getSubOrderId(), null);
            if (cnt > 0) {
                continue;
            }

            // 对应的套餐
            TaShopPackage shopPackage = shopPackageMapper.selectById(item.getPackageId());
            // 如果数量为2, 则生成2个
            for (int i = 0; i < item.getAmount(); i += 1) {
                TaVerifyTarget taVerifyTarget = new TaVerifyTarget();
                taVerifyTarget.setVerifyNo(getVerifyNo());
                taVerifyTarget.setTargetType(Constants.TARGET_SHOP_PACKAGE);
                taVerifyTarget.setTargetId(item.getPackageId());
                taVerifyTarget.setDescription(shopPackage.getDescription());
                taVerifyTarget.setPoster(shopPackage.getPoster());
                taVerifyTarget.setActualPrice(item.getUnitPrice());
                taVerifyTarget.setStandPrice(shopPackage.getStandardPrice());
                taVerifyTarget.setCashback(item.getCashback() / item.getAmount());
                taVerifyTarget.setCommissionCharge(item.getCommissionCharge() / item.getAmount());
                taVerifyTarget.setCreateDate(now);
                taVerifyTarget.setExpireDate(shopPackage.getEndTime());
                taVerifyTarget.setOrderDate(item.getCreateDate());
                taVerifyTarget.setOrderId(item.getOrderId());
                taVerifyTarget.setSubOrderId(item.getSubOrderId());
                taVerifyTarget.setShopId(item.getShopId());
                taVerifyTarget.setPersonId(taPerson.getPersonId());
                taVerifyTarget.setIsVerified(false);
                taVerifyTarget.setIsConfirmVerified(false);
                taVerifyTarget.setStatus(Constants.STATUS_NORMAL);

                taVerifyTargetMapper.insert(taVerifyTarget);
            }
        }
    }

    private String getVerifyNo() {
        while (true) {
            String verifyNo = StringUtils.random(10);
            int i = taVerifyTargetMapper.existsNo(verifyNo);
            if (i <1 ) {
                return verifyNo;
            }
        }
    }

    @Override
    public void refundNotifyOrder(WxPayRefundNotifyResult result) throws Exception {
        WxPayRefundNotifyResult.ReqInfo reqInfo = result.getReqInfo();
        if (null == reqInfo) {
            return;
        }

        // 退款处理中
        if (WxPayConstants.RefundStatus.PROCESSING.equals(reqInfo.getRefundStatus())) {
            return;
        }

        // 退款成功
        boolean isSuccess = WxPayConstants.RefundStatus.SUCCESS.equals(reqInfo.getRefundStatus());

        TaOrderPay orderPay = getById(reqInfo.getOutTradeNo());
        if ("返现".equals(orderPay.getDescription())) {
            cashback(orderPay);
            return;
        }


        orderPay.setPayDate(DateUtils.from(reqInfo.getSuccessTime(), "yyyyMMddHHmmss"));
        orderPay.setStatus(isSuccess ? Constants.PAY_SUCCESS : Constants.PAY_FAIL);
        orderPay.setMessage(isSuccess ? null : reqInfo.getRefundStatus());

        orderPayMapper.updateById(orderPay);

        String orderId = orderPay.getOrderId();
        String orderSubId = orderPay.getSubOrderId();
        if (isSuccess) {
            // 状态在申请退款的时候更新
//            if (StringUtils.isEmpty(orderSubId)) {
//                taOrderMapper.updateStatus(orderId, Constants.ORDER_REFUND);
//                taOrderSubMapper.updateStatusByOrder(orderId, Constants.ORDER_REFUND);
//            } else {
//                taOrderSubMapper.updateStatus(orderSubId, Constants.ORDER_REFUND);
//            }

            List<TaOrderSub> orderSubList = new ArrayList<>();
            if (null != orderId) {
                orderSubList = taOrderSubMapper.getListByOrder(orderId);
            } else {
                TaOrderSub taOrderSub = taOrderSubMapper.selectById(orderSubId);
                orderSubList.add(taOrderSub);
            }

            // 营业额处理
            dealSaleCharges(orderSubList, true);

            RefundSuccessEvent evt = new RefundSuccessEvent(this, orderPay);
            eventPublisher.publishEvent(evt);
        }
    }

    private void cashback(TaOrderPay taOrderPay) throws Exception {
        String orderId = taOrderPay.getOrderId();
        String orderSubId = taOrderPay.getSubOrderId();
        TaOrder taOrder = taOrderMapper.selectById(orderId);
        Integer cashback = taOrderPay.getCharges();
        boolean isOrderSub = !StringUtils.isEmpty(orderSubId);

        TaPersonAccount personAccount = personAccountMapper.getByPerson(taOrder.getPersonId());
        personAccountMapper.updateCashBack(personAccount.getAccountId(), cashback);

        TaPersonAccountLog accountLog = new TaPersonAccountLog();
        accountLog.setAccountId(personAccount.getAccountId());
        accountLog.setPersonId(taOrder.getPersonId());
        accountLog.setAmountType(Constants.ACCOUNT_CASHBACK);
        accountLog.setAmount(cashback);
        accountLog.setSignType(Constants.ACCOUNT_INCOME);
        accountLog.setTargetType(isOrderSub ? Constants.TARGET_ORDER_SUB : Constants.TARGET_ORDER);
        accountLog.setTargetId(isOrderSub ? orderSubId : orderId);
        accountLog.setOrderId(orderId);
        accountLogMapper.insert(accountLog);
    }

    @Override
    public WxPayMpOrderResult prePay(TaPerson person, TaOrder taOrder) throws Exception {
        TaOrderPay orderPay = createByOrder(taOrder);
        orderPayMapper.insert(orderPay);

        return wxPayUtils.createOrder(person, orderPay);
    }

    @Override
    public void refundOrderSub(TaPerson taPerson, TaOrder taOrder, TaOrderSub taOrderSub) throws Exception {
        TaOrderPay originOrderPay = null;

        // 找当前子项目对应的支付记录
        originOrderPay = orderPayMapper.getLastRowByOrderSub(taOrder.getOrderId(), taOrderSub.getSubOrderId());
        if (null != originOrderPay) {
            if (Constants.PAY_SUCCESS == originOrderPay.getStatus()) {
                throw new Exception("未找到当前订单有效的支付记录");
            }
        } else {
            // 找到主订单对应的那个支付记录
            originOrderPay = orderPayMapper.getLastPayedRowByOrder(taOrder.getOrderId());
            if (null != originOrderPay) {
                if (Constants.PAY_SUCCESS == originOrderPay.getStatus()) {
                    throw new Exception("未找到当前订单有效的支付记录");
                }
            } else {
                throw new Exception("未找到当前订单有效的支付记录");
            }
        }

        TaOrderPay orderPay = createByOrderSub(taOrderSub, taPerson);
        // 退款金额 = 订单金额 - 返现金额
        orderPay.setCharges(taOrderSub.getCharges() - taOrderSub.getCashback());
        orderPay.setIsRefund(true);

        // 发起退款
        wxPayUtils.refund(originOrderPay, orderPay);
    }

    @Override
    public void refundOrder(TaPerson person, TaOrder taOrder, Integer totalCashback) throws Exception {
        // 找到主订单对应的那个支付记录
        TaOrderPay originOrderPay = orderPayMapper.getLastPayedRowByOrder(taOrder.getOrderId());
        if (null != originOrderPay) {
            if (Constants.PAY_SUCCESS != originOrderPay.getStatus()) {
                throw new Exception("未找到当前订单有效的支付记录");
            }
        } else {
            throw new Exception("未找到当前订单有效的支付记录");
        }

        TaOrderPay orderPay = createByOrder(taOrder);
        // 退款金额 = 订单金额 - 返现金额
        orderPay.setIsRefund(true);
        orderPay.setCharges(taOrder.getCharges() - totalCashback);
        orderPayMapper.insert(orderPay);

        // 发起退款
        wxPayUtils.refund(originOrderPay, orderPay);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public WxPayMpOrderResult prePay(TaPerson person, TaOrder taOrder, TaOrderSub taOrderSub) throws Exception {
        TaOrderPay orderPay = createByOrderSub(taOrderSub, person);
        orderPayMapper.insert(orderPay);

        return wxPayUtils.createOrder(person, orderPay);
    }

    /**
     * 支付成功消息
     * @param taOrderPay
     */
    @Async
    @Override
    public void sendPaySuccessNotifyToShop(TaOrderPay taOrderPay) {
        try {
            // 消息模板
            TaMiniappMessageTpl messageTpl = messageTplMapper.selectByType(Constants.TPL_MESSAGE_SHOP_PAY_SUCCESS);
            if (null == messageTpl) {
                return ;
            }

            String templateId = messageTpl.getTemplateId();
            String page = messageTpl.getPage();

            String orderId = taOrderPay.getOrderId();
            String subOrderId = taOrderPay.getSubOrderId();

            List<Map<String, Object>> list = orderPayMapper.getShopWorkerBy(orderId, subOrderId);
            if (null == list || list.size() == 0) return;

            for (Map<String, Object> item : list) {
                String openId = item.get("openid").toString();
                if (StringUtils.isEmpty(openId)) continue;

                List<WxMaSubscribeMessage.Data> dataList = new ArrayList<>();
                dataList.add(new WxMaSubscribeMessage.Data("character_string7", orderId));
                dataList.add(new WxMaSubscribeMessage.Data("thing13", "您有一个十公里订单，客人期待与你的见面哦"));

                try {
                   wxUtils.sendMessage(templateId, openId, page, dataList);
                } catch (Exception e) {
                    e.printStackTrace();
                    log.error("发送订单消息失败: {}", e.getMessage());
                } finally {
                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private TaOrderPay createByOrder(TaOrder taOrder) throws Exception {
        // 发给微信的订单号为  订单ID-01 这种
        Integer no = taSequenceMapper.nextval(taOrder.getOrderId());
        String payNo = StringUtils.lpad(no.toString(), "0", 2);
        String payId = taOrder.getOrderId() + "-" + payNo;

        TaOrderPay orderPay = new TaOrderPay();
        orderPay.setPayId(payId);
        orderPay.setOrderId(taOrder.getOrderId());
        orderPay.setPersonId(taOrder.getPersonId());
        orderPay.setCharges(taOrder.getCharges());
        orderPay.setDescription(taOrder.getDescription());
        orderPay.setStatus(Constants.STATUS_READY);

        return orderPay;
    }

    private TaOrderPay createByOrderSub(TaOrderSub taOrderSub, TaPerson person) throws Exception {
        // 发给微信的订单号为  订单ID-01 这种
        Integer no = taSequenceMapper.nextval(taOrderSub.getSubOrderId());
        String payNo = StringUtils.lpad(no.toString(), "0", 2);
        String payId = taOrderSub.getSubOrderId() + "-" + payNo;

        TaOrderPay orderPay = new TaOrderPay();
        orderPay.setPayId(payId);
        orderPay.setOrderId(taOrderSub.getOrderId());
        orderPay.setSubOrderId(taOrderSub.getSubOrderId());
        orderPay.setPersonId(person.getPersonId());
        orderPay.setCharges(taOrderSub.getCharges());
        orderPay.setDescription(taOrderSub.getDescription());
        orderPay.setStatus(Constants.STATUS_READY);

        return orderPay;
    }
}
